# Third Party
from ninja import Schema

# Django
from django.utils import timezone


class CharacterSchema(Schema):
    character_id: int | None = None
    character_name: str | None = None
    character_portrait: str | None = None
    corporation_id: int | None = None
    corporation_name: str | None = None


class CorporationSchema(Schema):
    corporation_id: int | None = None
    corporation_name: str | None = None
    corporation_logo: str | None = None


class OreSchema(Schema):
    portrait: str | None = None
    name: str
    units: int
    volume_m3: int
    price_isk: float
    price_compressed: float | None = None
    timestamp: timezone.datetime
    snapshot: str
    html: str | None = None


class BeltSurveySessionSchema(Schema):
    public_id: str
    name: str
    created_at: timezone.datetime
    owner: str
    html: str | None = None


class OreSchemaList(Schema):
    snapshot: str | None = None
    timestamp: timezone.datetime | None = None
    entries: list[OreSchema]
    delete_html: str | None = None


class SurveyStatsSchema(Schema):
    size: float = 0.0
    left: float = 0.0
    mined: float = 0.0
    duration: float = 0.0
    start: timezone.datetime | None = None
    end: timezone.datetime | None = None
    rate: float = 0.0
    total_asteroids: int = 0
    remaining_asteroids: int = 0
    finish: timezone.datetime | None = None
    progress_percent: float = 0.0



"""Template for a TestView class."""

# Standard Library
from http import HTTPStatus

# Django
from django.urls import reverse

# AA Belt Radar
# AA Beltradar
from beltradar.tests import BeltRadarTestCase
from beltradar.tests.testdata.utils import create_user_from_evecharacter
from beltradar.views import index


class TestViews(BeltRadarTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

    def test_view(self):
        request = self.factory.get(reverse("beltradar:index"))
        request.user = self.user
        response = index(request)
        self.assertEqual(response.status_code, HTTPStatus.OK)



"""App URLs"""

# Django
from django.urls import path, re_path

# AA Belt Radar
from beltradar import views
from beltradar.api import api

app_name: str = "beltradar"  # pylint: disable=invalid-name

urlpatterns = [
    path("", views.index, name="index"),
    path("survey/create-session/", views.create_session, name="create_session"),
    path("survey/my-sessions/", views.view_my_sessions, name="view_my_sessions"),
    path(
        "survey/my-sessions/<int:character_id>/",
        views.view_my_sessions,
        name="view_my_sessions",
    ),
    path("survey/<uuid:public_id>/add-entry/", views.add_entry, name="add_entry"),
    path("survey/<uuid:public_id>/view/", views.view_session, name="view_session"),
    # -- API System
    re_path(r"^api/", api.urls),
]



"""App Configuration"""

# Django
from django.apps import AppConfig

# AA Belt Radar
from beltradar import __version__


class BeltRadarConfig(AppConfig):
    """App Config"""

    default_auto_field = "django.db.models.AutoField"
    name = "beltradar"
    label = "beltradar"
    verbose_name = f"Belt Radar v{__version__}"



# Standard Library
import socket
from unittest.mock import Mock

# Django
from django.contrib.messages.middleware import MessageMiddleware
from django.contrib.sessions.middleware import SessionMiddleware
from django.core.handlers.wsgi import WSGIRequest
from django.test import RequestFactory, TestCase

# AA Belt Radar
from beltradar.tests.testdata.integrations.allianceauth import load_allianceauth
from beltradar.tests.testdata.utils import create_user_from_evecharacter


class SocketAccessError(Exception):
    """Error raised when a test script accesses the network"""


class NoSocketsTestCase(TestCase):
    """Variation of Django's TestCase class that prevents any network use.

    Example:

        .. code-block:: python

            class TestMyStuff(BaseTestCase):
                def test_should_do_what_i_need(self): ...

    """

    @classmethod
    def setUpClass(cls):
        cls.socket_original = socket.socket
        socket.socket = cls.guard
        return super().setUpClass()

    @classmethod
    def tearDownClass(cls):
        socket.socket = cls.socket_original
        return super().tearDownClass()

    @staticmethod
    def guard(*args, **kwargs):
        raise SocketAccessError("Attempted to access network")


class BeltRadarTestCase(NoSocketsTestCase):
    """
    Preloaded Testcase class for Belt Radar tests without Network access.

    Pre-Load:
        * Alliance Auth Characters, Corporation, Alliance Data
        * Taken User IDs: 1001

    Available Request Factory:
        `self.factory`

    Available test users:
        * `user` User with standard Belt Radar access.
            * 'beltradar.basic_access' Permission
            * Character ID 1001
            * Corporation ID 2001
            * Alliance ID 3001

    Example:
        .. code-block:: python

            class TestMyBeltRadarStuff(BeltRadarTestCase):
                def test_should_do_what_i_need(self):
                    user = self.user
    """

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        # Initialize Alliance Auth test data
        load_allianceauth()

        # Request Factory
        cls.factory = RequestFactory()

        # User with Standard Access - Corporation 2001
        cls.user, cls.user_character = create_user_from_evecharacter(
            character_id=1001,
            permissions=["beltradar.basic_access"],
        )

    def _middleware_process_request(self, request: WSGIRequest):
        """Helper method to process middleware for a request."""
        session_middleware = SessionMiddleware(Mock())
        session_middleware.process_request(request)
        message_middleware = MessageMiddleware(Mock())
        message_middleware.process_request(request)



# Third Party
from ninja import NinjaAPI

# Django
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# AA Belt Radar
from beltradar import __title__
from beltradar.api import schema
from beltradar.api.helpers.core import get_owner_or_none, get_public_id_or_none
from beltradar.api.helpers.icons import (
    get_snapshot_delete_button,
    get_survey_manage_action_icons,
)
from beltradar.helpers.eveonline import get_icon_render_url
from beltradar.models.beltradar import BeltSurveySession
from beltradar.providers import AppLogger

logger = AppLogger(get_extension_logger(__name__), __title__)


class BeltRadarSurveyApiEndpoints:
    tags = ["Survey"]

    def __init__(self, api: NinjaAPI):
        @api.get(
            "view/session/{public_id}/",
            response={200: schema.BeltSurveySessionSchema, 403: str},
            tags=self.tags,
        )
        def get_survey_session(request, public_id: str):
            """Get survey session details for a specific public_id."""
            if not request.user.has_perm("beltradar.basic_access"):
                return 403, _("You do not have permission to access this resource.")

            session = BeltSurveySession.objects.filter(public_id=public_id).first()
            if not session:
                return 403, _("Survey session not found or not public.")

            survey_session_data = schema.BeltSurveySessionSchema(
                public_id=str(session.public_id),
                name=session.name,
                created_at=session.created_at,
                owner=str(session.owner),
                html=str(
                    get_survey_manage_action_icons(request=request, public_id=public_id)
                ),
            )
            return 200, survey_session_data

        @api.get(
            "view/my-sessions/{character_id}/",
            response={200: list[schema.BeltSurveySessionSchema], 403: dict, 404: dict},
            tags=self.tags,
        )
        def get_my_sessions(request, character_id: int):
            """Get all survey sessions for the current user."""
            perms = get_owner_or_none(request=request, character_id=character_id)[0]

            if perms is False:
                return 403, {"error": _("Permission Denied.")}

            sessions = BeltSurveySession.objects.filter(
                owner__profile__main_character__character_id=character_id
            ).order_by("-created_at")
            survey_list: list[schema.BeltSurveySessionSchema] = []
            for session in sessions:
                survey_session_data = schema.BeltSurveySessionSchema(
                    public_id=str(session.public_id),
                    name=session.name,
                    created_at=session.created_at,
                    owner=str(session.owner),
                    html=str(
                        get_survey_manage_action_icons(
                            request=request, public_id=session.public_id
                        )
                    ),
                )
                survey_list.append(survey_session_data)
            return 200, survey_list

        @api.get(
            "view/session/{public_id}/entries/",
            response={200: list[schema.OreSchemaList], 403: str},
            tags=self.tags,
        )
        def get_survey_entries(request, public_id: str):
            """Get all survey entries for the current user."""
            if not request.user.has_perm("beltradar.basic_access"):
                return 403, _("You do not have permission to access this resource.")

            session = BeltSurveySession.objects.filter(public_id=public_id).first()
            if not session:
                return 403, _("Survey session not found or not public.")

            survey_entries = session.br_entries.select_related(
                "eve_type"
            ).grouped_by_time()
            # Create a list of survey entries for the current user, ordered by timestamp
            survey_list: list[schema.OreSchemaList] = []
            for entries in survey_entries:
                ore_list = []
                for entry in entries:
                    ore_data = schema.OreSchema(
                        portrait=(
                            get_icon_render_url(type_id=entry.eve_type.id, as_html=True)
                            if entry.eve_type
                            else None
                        ),
                        name=entry.eve_type.name if entry.eve_type else "Unknown",
                        units=entry.units or 0,
                        volume_m3=entry.volume_left or 0,
                        price_isk=entry.price or 0.0,
                        price_compressed=entry.price_compressed or None,
                        timestamp=entry.timestamp,
                        snapshot=entry.snapshot,
                    )
                    ore_list.append(ore_data)
                survey_list.append(
                    schema.OreSchemaList(
                        snapshot=entries[0].snapshot,
                        timestamp=entries[0].timestamp,
                        entries=ore_list,
                        delete_html=str(
                            get_snapshot_delete_button(
                                request=request,
                                public_id=public_id,
                                snapshot=entries[0].snapshot,
                            )
                        ),
                    )
                )
            return 200, survey_list

        @api.get(
            "view/session/{public_id}/snapshot/last_entry/",
            response={200: schema.OreSchemaList, 403: str},
            tags=self.tags,
        )
        def get_survey_entry(request, public_id: str):
            """Get all survey entries for the current user."""
            if not request.user.has_perm("beltradar.basic_access"):
                return 403, _("You do not have permission to access this resource.")

            # Check if the survey session exists
            session = BeltSurveySession.objects.filter(public_id=public_id).first()
            if not session:
                return 403, _("Survey session not found or not public.")

            # Get the most recent survey entry for this session (if any)
            entries = (
                session.br_entries.select_related("eve_type")
                .order_by("-timestamp")
                .filter(snapshot=session.last_entry_snapshot())
            )

            # Create a list of survey entries for the current user, ordered by timestamp
            snapshot_list: list[schema.OreSchema] = []
            for entry in entries:
                ore_data = schema.OreSchema(
                    portrait=(
                        get_icon_render_url(type_id=entry.eve_type.id, as_html=True)
                        if entry.eve_type
                        else None
                    ),
                    name=entry.eve_type.name if entry.eve_type else "Unknown",
                    units=entry.units or 0,
                    volume_m3=entry.volume_left or 0,
                    price_isk=entry.price or 0.0,
                    price_compressed=entry.price_compressed or None,
                    timestamp=entry.timestamp,
                    snapshot=entry.snapshot,
                )
                snapshot_list.append(ore_data)
            return 200, schema.OreSchemaList(
                snapshot=entries[0].snapshot,
                timestamp=entries[0].timestamp,
                entries=snapshot_list,
                delete_html=str(
                    get_snapshot_delete_button(
                        request=request,
                        public_id=entries[0].session.public_id,
                        snapshot=entries[0].snapshot,
                    )
                ),
            )

        @api.post(
            "session/{public_id}/manage/delete/",
            response={200: dict, 403: dict, 404: dict},
            tags=self.tags,
        )
        def delete_survey_session(request, public_id: str):
            """
            Delete an entire survey session and all its entries.

            This Endpoint allows users to delete an entire survey session along with all its associated entries.
            The user must have permission to delete the survey session, and the survey session must exist.

            Args:
                public_id (str): The public UUID of the survey session.
            Returns:
                200: A success message indicating the survey session was deleted.
                403: An error message if the user does not have permission or the session is not found.
                404: An error message if no entries are found for the given survey session.
            """
            # Check if the survey session exists
            session = BeltSurveySession.objects.filter(public_id=public_id).first()
            if not session:
                msg = _("Survey session not found.")
                return 404, {"error": msg}

            # Check if the user has permission to delete this survey session
            perms = get_public_id_or_none(
                request=request,
                public_id=public_id,
            )[0]
            if not perms:
                msg = _("Permission Denied.")
                return 403, {"error": msg}

            # Delete the survey session and all associated entries
            session.delete()
            # If the session was deleted successfully, return a success message
            msg = _("Survey session and all associated entries deleted successfully.")
            return 200, {"success": True, "message": msg}

        @api.post(
            "session/{public_id}/snapshot/{snapshot}/manage/delete/",
            response={200: dict, 403: dict, 404: dict},
            tags=self.tags,
        )
        def delete_snapshot(request, public_id: str, snapshot: str):
            """
            Delete all survey entries for a specific snapshot in a survey session.

            This Endpoint allows users to delete all entries associated with a specific snapshot timestamp within a survey session.
            The user must have permission to delete the survey session, and the survey session must exist.

            Args:
                public_id (str): The public UUID of the survey session.
                snapshot (str): The timestamp of the snapshot to delete, in ISO format.
            Returns:
                200: A success message indicating the snapshot was deleted.
                403: An error message if the user does not have permission or the session is not found.
                404: An error message if no entries are found for the given snapshot.
            """
            # Check if the survey session exists
            session = BeltSurveySession.objects.filter(public_id=public_id).first()
            if not session:
                msg = _("Survey session not found.")
                return 404, {"error": msg}

            # Check if the user has permission to delete this snapshot (by checking if they can delete the survey session)
            perms = get_owner_or_none(
                request=request,
                character_id=session.owner.profile.main_character.character_id,
            )[0]
            if not perms:
                msg = _("Permission Denied.")
                return 403, {"error": msg}

            deleted_count = session.br_entries.filter(snapshot=snapshot).delete()[0]
            if deleted_count == 0:
                msg = _("No entries found for the given snapshot.")
                return 404, {"error": msg}

            # If entries were deleted successfully, return a success message
            msg = _("Snapshot deleted successfully.")
            return 200, {"success": True, "message": msg}



# Django
from django.core.exceptions import ObjectDoesNotExist

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# AA Belt Radar
from beltradar import __title__, models
from beltradar.providers import AppLogger

logger = AppLogger(get_extension_logger(__name__), __title__)


def get_owner_or_none(
    request, character_id
) -> tuple[bool, models.BeltSurveySession | None]:
    """Get Character and check permissions"""
    perms = True

    try:
        survey_session = models.BeltSurveySession.objects.filter(
            owner__profile__main_character__character_id=character_id
        ).first()
    except ObjectDoesNotExist:
        return False, None
    except ValueError:
        return None, None

    # check access
    visible = models.BeltSurveySession.objects.visible_eve_characters(request.user)
    if survey_session.owner.profile.main_character not in visible:
        perms = False
    return perms, survey_session


def get_public_id_or_none(
    request, public_id
) -> tuple[bool, models.BeltSurveySession | None]:
    """Get Character and check permissions"""
    perms = True

    try:
        survey_session = models.BeltSurveySession.objects.get(public_id=public_id)
    except ObjectDoesNotExist:
        return False, None
    except ValueError:
        return None, None

    # check access
    visible = models.BeltSurveySession.objects.visible_eve_characters(request.user)
    if survey_session.owner.profile.main_character not in visible:
        perms = False
    return perms, survey_session



# Generated by Django 5.2.13 on 2026-05-16 14:23

# Standard Library
import uuid

# Django
import django.db.models.deletion
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("eve_sde", "0018_blueprintactivity_blueprintactivityproduct_and_more"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="General",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            options={
                "permissions": (
                    ("basic_access", "Can access this app, Belt Radar"),
                    ("manage_access", "Can manage Belt Radar"),
                    ("admin_access", "Admin access to Belt Radar"),
                ),
                "managed": False,
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="BeltSurveySession",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "public_id",
                    models.UUIDField(default=uuid.uuid4, editable=False, unique=True),
                ),
                ("name", models.CharField(blank=True, max_length=150, null=True)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                (
                    "owner",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="br_user_sessions",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="BeltSurveyEntry",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("snapshot", models.CharField(blank=True, max_length=64, null=True)),
                ("timestamp", models.DateTimeField(default=django.utils.timezone.now)),
                ("units", models.IntegerField(blank=True, null=True)),
                ("volume_left", models.FloatField(blank=True, null=True)),
                ("note", models.TextField(blank=True, null=True)),
                ("price_compressed", models.FloatField(blank=True, null=True)),
                ("price", models.FloatField(blank=True, null=True)),
                (
                    "eve_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="br_survey_entries",
                        to="eve_sde.itemtype",
                    ),
                ),
                (
                    "recorded_by",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "session",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="br_entries",
                        to="beltradar.beltsurveysession",
                    ),
                ),
            ],
            options={
                "ordering": ["timestamp"],
                "default_permissions": (),
            },
        ),
    ]



"""This module provides helper functions to get EVE Online related images and renders."""

# Django
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.eveonline.evelinks.eveimageserver import (
    alliance_logo_url,
    character_portrait_url,
    corporation_logo_url,
    type_icon_url,
    type_render_url,
)


def get_character_portrait_url(
    character_id: int, size: int = 32, character_name: str = "", as_html: bool = False
) -> str:
    """
    Get the character portrait for a character ID.

    Args:
        character_id (int): The ID of the character.
        size (int, optional): The size of the portrait image.
        character_name (str, optional): The name of the character.
        as_html (bool, optional): Whether to return the portrait as an HTML img tag.
    Returns:
        str: The URL of the character portrait or an HTML img tag.
    """
    try:
        render_url = character_portrait_url(character_id=character_id, size=size)
    except ValueError:
        return ""

    if as_html:
        render_html = format_html(
            '<img class="character-portrait rounded-circle" src="{}" alt="{}">',
            render_url,
            character_name,
        )
        return render_html
    return render_url


def get_corporation_logo_url(
    corporation_id: int,
    size: int = 32,
    corporation_name: str = "",
    as_html: bool = False,
) -> str:
    """
    Get the corporation logo for a corporation ID.

    Args:
        corporation_id (int): The ID of the corporation.
        size (int, optional): The size of the logo image.
        corporation_name (str, optional): The name of the corporation.
        as_html (bool, optional): Whether to return the logo as an HTML img tag.
    Returns:
        str: The URL of the corporation logo or an HTML img tag.
    """

    render_url = corporation_logo_url(corporation_id=corporation_id, size=size)

    if as_html:
        render_html = format_html(
            '<img class="corporation-logo rounded-circle" src="{}" alt="{}">',
            render_url,
            corporation_name,
        )
        return render_html
    return render_url


def get_alliance_logo_url(
    alliance_id: int,
    size: int = 32,
    alliance_name: str = "",
    as_html: bool = False,
) -> str:
    """
    Get the alliance logo for a alliance ID.

    Args:
        alliance_id (int): The ID of the alliance.
        size (int, optional): The size of the logo image.
        alliance_name (str, optional): The name of the alliance.
        as_html (bool, optional): Whether to return the logo as an HTML img tag.

    Returns:
        str: The URL of the alliance logo or an HTML img tag.
    """

    render_url = alliance_logo_url(alliance_id=alliance_id, size=size)

    if as_html:
        render_html = format_html(
            '<img class="alliance-logo rounded-circle" src="{}" alt="{}">',
            render_url,
            alliance_name,
        )
        return render_html
    return render_url


def get_type_render_url(
    type_id: int, size: int = 32, type_name: str = "", as_html: bool = False
) -> str:
    """
    Get the type render for a type ID.

    Args:
        type_id (int): The ID of the type.
        size (int, optional): The size of the render image.
        type_name (str, optional): The name of the type.
        as_html (bool): Whether to return the render as an HTML img tag.

    Returns:
        str: The URL of the type render or an HTML img tag.
    """

    render_url = type_render_url(type_id=type_id, size=size)

    if as_html:
        render_html = format_html(
            '<img class="type-render rounded-circle" src="{}" alt="{}">',
            render_url,
            type_name,
        )
        return render_html
    return render_url


def get_icon_render_url(
    type_id: int, size: int = 32, type_name: str = "", as_html: bool = False
) -> str:
    """
    Get the icon render for a type ID.

    Args:
        type_id (int): The ID of the type.
        size (int, optional): The size of the icon image.
        type_name (str, optional): The name of the type.
        as_html (bool): Whether to return the icon as an HTML img tag.

    Returns:
        str: The URL of the type icon or an HTML img tag.
    """

    render_url = type_icon_url(type_id=type_id, size=size)

    if as_html:
        render_html = format_html(
            '<img class="type-render rounded-circle" data-bs-tooltip="aa-ledger" src="{}" title="{}">',
            render_url,
            type_name,
        )
        return render_html
    return render_url



"""Models for Belt Radar."""

# Standard Library
import uuid
from collections import defaultdict

# Django
from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.authentication.models import User
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from eve_sde.models import ItemType

# AA Belt Radar
from beltradar import __title__
from beltradar.managers import BeltRadarManager, BeltSurveyEntryManager
from beltradar.providers import AppLogger

logger = AppLogger(get_extension_logger(__name__), __title__)


class BeltSurveySession(models.Model):
    """Represents a single survey session for a belt, which can have multiple entries (BeltSurveyEntry)."""

    class Meta:
        default_permissions = ()  # Remove standard permissions

    objects: BeltRadarManager = BeltRadarManager()

    owner = models.ForeignKey(
        User, on_delete=models.CASCADE, related_name="br_user_sessions"
    )
    public_id = models.UUIDField(default=uuid.uuid4, editable=False, unique=True)
    name = models.CharField(max_length=150, blank=True, null=True)
    created_at = models.DateTimeField(auto_now_add=True)

    # pylint: disable=too-many-locals
    def mining_stats(self):
        """
        Unified mining statistics for this session.

        Returns a SurveyStatsSchema with aggregated data from all entries in this session.
        """
        # pylint: disable=import-outside-toplevel
        # AA Belt Radar
        from beltradar.api import schema

        entries = list(self.br_entries.select_related("eve_type").order_by("timestamp"))
        if len(entries) < 2:  # Not enough data points to calculate meaningful stats
            return schema.SurveyStatsSchema()

        snapshots: dict = defaultdict(float)
        snapshots_entries: dict[timezone.datetime, list[float]] = defaultdict(list)
        for entry in entries:
            try:
                vol = float(getattr(entry, "volume_left", None))
            except (TypeError, ValueError):
                vol = None
            if vol is None:
                continue
            ts = entry.timestamp.replace(microsecond=0)
            snapshots[ts] += vol
            snapshots_entries[ts].append(vol)

        # If we have no valid snapshots, return empty stats
        if not snapshots:
            return schema.SurveyStatsSchema()

        sorted_ts = sorted(snapshots.keys())
        start_ts, end_ts = sorted_ts[0], sorted_ts[-1]
        size, size_left = snapshots[start_ts], snapshots[end_ts]
        size_mined = max(0.0, size - size_left)
        duration = (end_ts - start_ts).total_seconds()

        vols_start = snapshots_entries.get(start_ts, [])
        vols_last = snapshots_entries.get(end_ts, [])
        asteroids_total = (
            len(vols_start) if vols_start else (len(vols_last) if vols_last else None)
        )
        asteroids_left = sum(1 for v in vols_last if v > 0.0) if vols_last else None

        rate = 0.0
        if len(sorted_ts) >= 2:
            prev_vol = snapshots[sorted_ts[-2]]
            mined_delta = max(0.0, prev_vol - size_left)
            time_delta = (end_ts - sorted_ts[-2]).total_seconds()
            if time_delta > 0 and mined_delta > 0:
                rate = mined_delta / time_delta
        elif duration > 0 and size_mined > 0:
            rate = size_mined / duration

        eta_seconds = size_left / rate if rate > 0 else None
        finish_dt = None
        if eta_seconds is not None:
            try:
                finish_dt = end_ts + timezone.timedelta(seconds=int(eta_seconds))
            except (OverflowError, OSError):
                pass
        progress_percent = (
            min(100.0, max(0.0, (size_mined / size) * 100.0)) if size else 0.0
        )

        return schema.SurveyStatsSchema(
            size=size,
            left=size_left,
            mined=size_mined,
            duration=duration,
            start=start_ts,
            end=end_ts,
            rate=rate,
            total_asteroids=asteroids_total,
            remaining_asteroids=asteroids_left,
            finish=finish_dt,
            progress_percent=progress_percent,
        )

    def last_entry(self):
        """Get the most recent survey entry in this session, or None if there are no entries."""
        return self.br_entries.order_by("-timestamp").first()

    def last_entry_snapshot(self):
        """Get the snapshot identifier of the most recent entry in this session, or None if there are no entries."""
        last = self.last_entry()
        return last.snapshot if last else None

    @property
    def is_fresh(self):
        last = self.last_entry()
        if not last:
            return False
        age_seconds = (timezone.now() - last.timestamp).total_seconds()
        return (
            age_seconds < 300
        )  # consider fresh if last entry is less than 5 minutes old


class BeltSurveyEntry(models.Model):
    """Represents a single survey entry for a belt, linked to a BeltSurveySession."""

    objects: BeltSurveyEntryManager = BeltSurveyEntryManager()

    class Meta:
        default_permissions = ()  # Remove standard permissions
        ordering = ["timestamp"]

    session = models.ForeignKey(
        BeltSurveySession, on_delete=models.CASCADE, related_name="br_entries"
    )
    snapshot = models.CharField(max_length=64, null=True, blank=True)
    recorded_by = models.ForeignKey(
        User, on_delete=models.SET_NULL, null=True, blank=True
    )
    timestamp = models.DateTimeField(default=timezone.now)
    eve_type = models.ForeignKey(
        ItemType, on_delete=models.CASCADE, related_name="br_survey_entries"
    )
    units = models.IntegerField(null=True, blank=True)
    volume_left = models.FloatField(null=True, blank=True)
    note = models.TextField(null=True, blank=True)
    price_compressed = models.FloatField(null=True, blank=True)
    price = models.FloatField(null=True, blank=True)



# Standard Library
from datetime import date as dt_date
from typing import TYPE_CHECKING

# Django
from django.db import models

# Alliance Auth
from allianceauth.eveonline.models import EveCharacter
from allianceauth.services.hooks import get_extension_logger

# AA Belt Radar
from beltradar import __title__
from beltradar.providers import AppLogger

if TYPE_CHECKING:
    # AA Belt Radar
    from beltradar.models import BeltSurveyEntry as BeltSurveyEntryContext
    from beltradar.models import BeltSurveySession as BeltSurveyContext

logger = AppLogger(get_extension_logger(__name__), __title__)


class BeltRadarQuerySet(models.QuerySet["BeltSurveyContext"]):
    def visible_to(self, user):
        """Get all survey sessions visible to the user."""
        # superusers get all visible
        if user.is_superuser:
            logger.debug(
                "Returning all survey sessions for superuser %s.",
                user,
            )
            return self

        if user.has_perm("taxsystem.admin_access"):
            logger.debug("Returning all survey sessions for admin user %s.", user)
            return self

        try:
            char = user.profile.main_character
            assert char
            queries = [models.Q(eve_character__character_ownership__user=user)]

            logger.debug(
                "%s queries for user %s visible survey sessions.", len(queries), user
            )

            query = queries.pop()
            for q in queries:
                query |= q
            return self.filter(query)
        except AssertionError:
            logger.debug("User %s has no main character. Nothing visible.", user)
            return self.none()

    def manage_to(self, user):
        """Get all survey sessions that the user can manage."""
        # superusers get all visible
        if user.is_superuser:
            logger.debug(
                "Returning all survey sessions for superuser %s.",
                user,
            )
            return self

        if user.has_perm("taxsystem.admin_access"):
            logger.debug("Returning all survey sessions for admin user %s.", user)
            return self

        try:
            char = user.profile.main_character
            assert char
            query = models.Q(eve_character__character_ownership__user=user)
            logger.debug("Returning own survey sessions for User %s.", user)

            if query is None:
                return self.none()

            return self.filter(query)
        except AssertionError:
            logger.debug("User %s has no main character. Nothing visible.", user)
            return self.none()


class BeltRadarManager(models.Manager["BeltSurveyContext"]):
    def get_queryset(self) -> BeltRadarQuerySet:
        return BeltRadarQuerySet(self.model, using=self._db)

    def visible_to(self, user):
        return self.get_queryset().visible_to(user)

    def manage_to(self, user):
        return self.get_queryset().manage_to(user)

    @staticmethod
    def visible_eve_characters(user):
        qs = EveCharacter.objects.get_queryset()
        if user.is_superuser:
            logger.debug("Returning all characters for superuser %s.", user)
            return qs.all()

        if user.has_perm("beltradar.admin_access"):
            logger.debug("Returning all characters for %s.", user)
            return qs.all()

        try:
            char = user.profile.main_character
            assert char
            queries = [models.Q(character_ownership__user=user)]

            logger.debug(
                "%s queries for user %s visible chracters.", len(queries), user
            )

            query = queries.pop()
            for q in queries:
                query |= q
            return qs.filter(query)
        except AssertionError:
            logger.debug("User %s has no main character. Nothing visible.", user)
            return qs.none()


class BeltSurveyEntryQuerySet(models.QuerySet["BeltSurveyEntryContext"]):
    def grouped_by_time(
        self, date: dt_date | None = None, tolerance_seconds: int = 60
    ) -> list[list["BeltSurveyEntryContext"]]:
        """Group entries by timestamp proximity.

        Entries are ordered by timestamp and grouped when they are within
        `tolerance_seconds` from the first entry of the current group.
        """
        qs = self.order_by("timestamp")
        if date is not None:
            qs = qs.filter(timestamp__date=date)

        entries = list(qs)
        if not entries:
            return []

        groups: list[list["BeltSurveyEntryContext"]] = []
        current_group = [entries[0]]
        group_start = entries[0].timestamp

        for entry in entries[1:]:
            if (entry.timestamp - group_start).total_seconds() <= tolerance_seconds:
                current_group.append(entry)
                continue

            groups.append(current_group)
            current_group = [entry]
            group_start = entry.timestamp

        groups.append(current_group)
        return groups


class BeltSurveyEntryManager(models.Manager["BeltSurveyEntryContext"]):
    def get_queryset(self) -> BeltSurveyEntryQuerySet:
        return BeltSurveyEntryQuerySet(self.model, using=self._db)

    def grouped_by_time(
        self, date: dt_date | None = None, tolerance_seconds: int = 60
    ) -> list[list["BeltSurveyEntryContext"]]:
        return self.get_queryset().grouped_by_time(
            date=date,
            tolerance_seconds=tolerance_seconds,
        )



# Standard Library
import datetime as dt
import random
import string

# Django
from django.contrib.auth.models import User

# Alliance Auth
from allianceauth.authentication.backends import StateBackend
from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveCharacter
from allianceauth.tests.auth_utils import AuthUtils
from esi.models import Scope, Token


def dt_eveformat(my_dt: dt.datetime) -> str:
    """Convert datetime to EVE Online ISO format (YYYY-MM-DDTHH:MM:SS)

    Args:
        my_dt (datetime): Input datetime
    Returns:
        str: datetime in EVE Online ISO format
    """

    my_dt_2 = dt.datetime(
        my_dt.year, my_dt.month, my_dt.day, my_dt.hour, my_dt.minute, my_dt.second
    )

    return my_dt_2.isoformat()


def random_string(char_count: int) -> str:
    """returns a random string of given length"""
    return "".join(
        random.choice(string.ascii_uppercase + string.digits) for _ in range(char_count)
    )


def _generate_token(
    character_id: int,
    character_name: str,
    owner_hash: str | None = None,
    access_token: str = "access_token",
    refresh_token: str = "refresh_token",
    scopes: list | None = None,
    timestamp_dt: dt.datetime | None = None,
    expires_in: int = 1200,
) -> dict:
    """Generates the input to create a new SSO test token.

    Args:
        character_id (int): Character ID
        character_name (str): Character Name
        owner_hash (Optional[str], optional): Character Owner Hash. Defaults to None.
        access_token (str, optional): Access Token string. Defaults to "access_token".
        refresh_token (str, optional): Refresh Token string. Defaults to "refresh_token".
        scopes (Optional[list], optional): List of scope names. Defaults to None.
        timestamp_dt (Optional[dt.datetime], optional): Timestamp datetime. Defaults to None.
        expires_in (int, optional): Expiry time in seconds. Defaults to 1200
    Returns:
        dict: The generated token dict
    """
    if timestamp_dt is None:
        timestamp_dt = dt.datetime.utcnow()
    if scopes is None:
        scopes = [
            "esi-mail.read_mail.v1",
            "esi-wallet.read_character_wallet.v1",
            "esi-universe.read_structures.v1",
        ]
    if owner_hash is None:
        owner_hash = random_string(28)
    token = {
        "access_token": access_token,
        "token_type": "Bearer",
        "expires_in": expires_in,
        "refresh_token": refresh_token,
        "timestamp": int(timestamp_dt.timestamp()),
        "CharacterID": character_id,
        "CharacterName": character_name,
        "ExpiresOn": dt_eveformat(timestamp_dt + dt.timedelta(seconds=expires_in)),
        "Scopes": " ".join(list(scopes)),
        "TokenType": "Character",
        "CharacterOwnerHash": owner_hash,
        "IntellectualProperty": "EVE",
    }
    return token


def _store_as_Token(token: dict, user: object) -> Token:
    """Stores a generated token dict as Token object for given user

    Args:
        token (dict): Generated token dict
        user (User): Alliance Auth User
    Returns:
        Token: The created Token object
    """
    character_tokens = user.token_set.filter(character_id=token["CharacterID"])
    if character_tokens.exists():
        token["CharacterOwnerHash"] = character_tokens.first().character_owner_hash
    obj = Token.objects.create(
        access_token=token["access_token"],
        refresh_token=token["refresh_token"],
        user=user,
        character_id=token["CharacterID"],
        character_name=token["CharacterName"],
        token_type=token["TokenType"],
        character_owner_hash=token["CharacterOwnerHash"],
    )
    for scope_name in token["Scopes"].split(" "):
        scope, _ = Scope.objects.get_or_create(name=scope_name)
        obj.scopes.add(scope)
    return obj


def add_new_token(
    user: User,
    character: EveCharacter,
    scopes: list[str] | None = None,
    owner_hash: str | None = None,
) -> Token:
    """Generate a new token for a user based on a character and makes the given user it's owner.

    Args:
        user: Alliance Auth User
        character: EveCharacter to create the token for
        scopes: list of scope names
        owner_hash: optional owner hash to use for the token
    Returns:
        Token: The created Token object
    """
    return _store_as_Token(
        _generate_token(
            character_id=character.character_id,
            character_name=character.character_name,
            owner_hash=owner_hash,
            scopes=scopes,
        ),
        user,
    )


def add_character_to_user(
    user: User,
    character: EveCharacter,
    is_main: bool = False,
    scopes: list[str] | None = None,
    disconnect_signals: bool = False,
) -> CharacterOwnership:
    """Add an existing :class:`EveCharacter` to a User, optionally as main character.

    Args:
        user (User): The User to whom the EveCharacter will be added.
        character (EveCharacter): The EveCharacter to add to the User.
        is_main (bool, optional): Whether to set the EveCharacter as the User's main
            character. Defaults to ``False``.
        scopes (list[str] | None, optional): List of scope names to assign to the
            character's token. If ``None``, defaults to `["publicData"]`. Defaults to ``None``.
        disconnect_signals (bool, optional): Whether to disconnect signals during
            the addition. Defaults to ``False``.
    Returns:
        CharacterOwnership: The created CharacterOwnership instance.
    """

    if not scopes:
        scopes = ["publicData"]

    if disconnect_signals:
        AuthUtils.disconnect_signals()

    add_new_token(user, character, scopes)

    if is_main:
        user.profile.main_character = character
        user.profile.save()
        user.save()

    if disconnect_signals:
        AuthUtils.connect_signals()

    return CharacterOwnership.objects.get(user=user, character=character)


def create_user_from_evecharacter(
    character_id: int,
    permissions: list[str] | None = None,
    scopes: list[str] | None = None,
) -> tuple[User, CharacterOwnership]:
    """Create new allianceauth user from EveCharacter object.

    Args:
        character_id (int): ID of eve character
        permissions (list[str] | None): list of permission names, e.g. `"my_app.my_permission"`
        scopes (list[str] | None): list of scope names
    Returns:
        tuple(User, CharacterOwnership): Created Alliance Auth User and CharacterOwnership
    """
    auth_character = EveCharacter.objects.get(character_id=character_id)
    user = AuthUtils.create_user(auth_character.character_name.replace(" ", "_"))
    character_ownership = add_character_to_user(
        user, auth_character, is_main=True, scopes=scopes
    )
    if permissions:
        for permission_name in permissions:
            user = AuthUtils.add_permission_to_user_by_name(permission_name, user)
    return user, character_ownership


def add_permission_to_user(
    user: User,
    permissions: list[str] | None = None,
) -> User:
    """Add permission to existing allianceauth user.
    Args:
        user (User): Alliance Auth User
        permissions (list[str] | None): list of permission names, e.g. `"my_app.my_permission"`
    Returns:
        User: Updated Alliance Auth User
    """
    if permissions:
        for permission_name in permissions:
            user = AuthUtils.add_permission_to_user_by_name(permission_name, user)
            return user
    raise ValueError("No permissions provided to add to user.")



"""App Tasks"""

# Third Party
from celery import shared_task

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger
from allianceauth.services.tasks import QueueOnce

# AA Belt Radar
from beltradar import __title__, app_settings
from beltradar.models import BeltSurveySession
from beltradar.providers import AppLogger

logger = AppLogger(get_extension_logger(__name__), __title__)

MAX_RETRIES_DEFAULT = 3

# Default params for all tasks.
TASK_DEFAULTS = {
    "time_limit": app_settings.BELT_RADAR_TASK_TIME_LIMIT,
    "max_retries": MAX_RETRIES_DEFAULT,
}

# Default params for tasks that need bind=True.
TASK_DEFAULTS_BIND = {**TASK_DEFAULTS, **{"bind": True}}

# Default params for tasks that need bind=True and run once only.
TASK_DEFAULTS_BIND_ONCE = {**TASK_DEFAULTS, **{"bind": True, "base": QueueOnce}}

# Default params for tasks that need run once only.
TASK_DEFAULTS_ONCE = {**TASK_DEFAULTS, **{"base": QueueOnce}}

TASK_DEFAULTS_BIND_ONCE_OWNER = {
    **TASK_DEFAULTS_BIND_ONCE,
    **{"once": {"keys": ["eve_id"], "graceful": True}},
}


@shared_task(**TASK_DEFAULTS_ONCE)
def update_all_belt_radar(runs: int = 0, force_refresh: bool = False):
    """Update all belt radar data."""
    sessions = BeltSurveySession.objects.all()
    for session in sessions:
        update_belt_radar_session.apply_async(
            kwargs={"public_id": session.public_id, "force_refresh": force_refresh}
        )
        runs = runs + 1

    logger.info("Queued %s Session Tasks", runs)


# pylint: disable=unused-argument
@shared_task(**TASK_DEFAULTS_BIND_ONCE_OWNER)
def update_belt_radar_session(self, public_id: str, force_refresh: bool = False):
    """Update a single belt radar session."""
    # Placeholder for future implementation, currently no background processing needed for session updates.
